@@ -19,6 +19,7 @@
 #include <unistd.h>
 #include <sys/param.h>
 #include <dirent.h>
+#include <linux/kdev_t.h>
 
 /*
  * FIXME: really need to seperate names from the devices since
@@ -120,28 +121,6 @@ static int _insert_dev(const char *path, dev_t d)
 	return 1;
 }
 
-/*
- * return a new path for the destination of the link.
- */
-static char *_follow_link(const char *path, struct stat *info)
-{
-	char buffer[PATH_MAX + 1];
-	int n;
-	n = readlink(path, buffer, sizeof(buffer) - 1);
-
-	if (n <= 0)
-		return NULL;
-
-	buffer[n] = '\0';
-
-	if (stat(buffer, info) < 0) {
-		log_sys_very_verbose("stat", buffer);
-		return NULL;
-	}
-
-	return pool_strdup(_cache.mem, buffer);
-}
-
 static char *_join(const char *dir, const char *name)
 {
 	int len = strlen(dir) + strlen(name) + 2;
@@ -208,7 +187,6 @@ static int _insert_dir(const char *dir)
 static int _insert(const char *path, int rec)
 {
 	struct stat info;
-	char *actual_path;
 	int r = 0;
 
 	if (stat(path, &info) < 0) {
@@ -216,22 +194,6 @@ static int _insert(const char *path, int rec)
 		return 0;
 	}
 
-	/* follow symlinks if neccessary. */
-	if (S_ISLNK(info.st_mode)) {
-		log_debug("%s: Following symbolic link", path);
-		if (!(actual_path = _follow_link(path, &info))) {
-			stack;
-			return 0;
-		}
-
-		if (stat(actual_path, &info) < 0) {
-			log_sys_very_verbose("stat", actual_path);
-			return 0;
-		}
-
-		dbg_free(actual_path);
-	}
-
 	if (S_ISDIR(info.st_mode)) { /* add a directory */
 		if (rec)
 			r = _insert_dir(path);
@@ -306,7 +268,7 @@ void _check_closed(struct device *dev)
 
 static inline void _check_for_open_devices(void)
 {
-	hash_iterate(_cache.names, (iterate_fn)_check_closed);
+	hash_iter(_cache.names, (iterate_fn)_check_closed);
 }
 
 void dev_cache_exit(void)
@@ -342,10 +304,56 @@ int dev_cache_add_dir(const char *path)
 	return 1;
 }
 
+/* Check cached device name is still valid before returning it */
+/* This should be a rare occurrence */
+/* FIXME Make rest of code pass/cache struct device instead of dev_name */
+const char *dev_name_confirmed(struct device *dev)
+{
+	struct stat buf;
+	char *name;
+	int r;
+
+	while ((r = stat(name = list_item(dev->aliases.n, 
+					   struct str_list)->str, &buf)) || 
+	       (buf.st_rdev != dev->dev)) {
+		if (r < 0)
+			log_sys_error("stat", name);
+		log_error("Path %s no longer valid for device(%d,%d)",
+			  name, (int) MAJOR(dev->dev), (int) MINOR(dev->dev));
+
+		/* Remove the incorrect hash entry */
+		hash_remove(_cache.names, name);
+
+		/* Leave list alone if there isn't an alternative name */
+		/* so dev_name will always find something to return. */
+		/* Otherwise add the name to the correct device. */
+		if (list_size(&dev->aliases) > 1) {
+			list_del(dev->aliases.n);
+			if (!r)
+				_insert(name, 0);
+			continue;
+		}
+
+		log_error("Aborting - please provide new pathname for what "
+			  "used to be %s", name);
+		return NULL;
+	}
+
+	return dev_name(dev);
+}
+
+
 struct device *dev_cache_get(const char *name, struct dev_filter *f)
 {
+	struct stat buf;
 	struct device *d = (struct device *) hash_lookup(_cache.names, name);
 
+	/* If the entry's wrong, remove it */
+	if (d && (stat(name, &buf) || (buf.st_rdev != d->dev))) {
+		hash_remove(_cache.names, name);
+		d = NULL;
+	}
+
 	if (!d) {
 		_insert(name, 0);
 		d = (struct device *) hash_lookup(_cache.names, name);
@@ -392,5 +400,3 @@ struct device *dev_iter_get(struct dev_iter *iter)
 	return NULL;
 }
 
-
-
