@@ -21,7 +21,7 @@
 #ifndef _LVM_FILTER_H
 #define _LVM_FILTER_H
 
-struct dev_filter *lvm_type_filter_create();
+struct dev_filter *lvm_type_filter_create(const char *);
 
 void lvm_type_filter_destroy(struct dev_filter *f);
 
